/* exons2mrna.c -- use an exons file to extract mRNA from genomic sequence
 *
 * This program reads an exons file and a genomic sequence,
 * and writes out a FastA library of putative mRNA sequences.
 * CDS specifications are included when available, e.g.
 *
 *      >gene1 2328 bp CDS=143..2089
 *      ACGT...
 *      >gene2 1345 bp CDS=321..1234
 *      TGCA...
 *
 * $Revision: 1.18 $$Date: 2001/04/04 21:40:58 $
 */

#include "exons.h"
#include "seq.h"
#include "util.h"

#define SYNTAX  "exons_file genomic_seq_file > output_file"

enum { BUFSIZE = 1000000, WIDTH = 60 };

static void get_args (int argc, char **argv, char **exonfile, char **seqfile);
static void print_hdr (char *gname, int gdir, int length, int mcds_begin,
                       int mcds_end);
static void print_seq (int gdir, int length, uchar *mrna);

int main (int argc, char **argv) {
	char *exonfile, *seqfile;               /* names of input files         */
	char **gene_name, **exon_name;          /* arrays of gene/exon info     */
	int *begin, *end, *dir;                 /* arrays of gene/exon info     */
	int exlen;                              /* length of arrays             */
	SEQ *sp;                                /* ptr to struct holding seq    */
	uchar *seq;                             /* genomic sequence (start = 0) */
	int seqlen;                             /* length of genomic sequence   */
	char *gname;                            /* name of current gene         */
	int gdir;                               /* direction of current gene    */
	int gcds_begin, gcds_end;               /* translated region (genomic)  */
	int mcds_begin, mcds_end;               /* translated region (mrna)     */
	int xbeg, xend;                         /* exon endpoints               */
	uchar mrna [BUFSIZE];                   /* buffer for mrna sequence     */
	uchar *mptr;                            /* next position in buffer      */
	int i, j;                               /* indices                      */

	/* get arguments */
	set_argv0 (argv[0]);
	get_args (argc, argv, &exonfile, &seqfile);
	if (!exonfile || !seqfile)
		fatal (usage (SYNTAX));

	/* read exon info */
	exlen = get_exons (exonfile, &begin, &end, &dir, &gene_name,
	                   &exon_name, -1);    /* print all errors */

	/* read genomic sequence */
	if (!(sp = seq_get (seqfile)))
		fatalf ("Couldn't get sequence \"%s\".", seqfile);
	seq = SEQ_CHARS (sp);
	seqlen = SEQ_LEN (sp);
	if (SEQ_FROM (sp) != 1)
		fatal ("Specification of subsequence is not supported.");

	/* print mrna sequences */
	assert (!exlen || gene_name[0]);
	gname = NULL; mptr = NULL;    /* pacify lint */
	gdir = gcds_begin = gcds_end = mcds_begin = mcds_end = 0;
	for (i = 0; i < exlen; i++) {
		if (gene_name[i]) {     /* new gene */
			/* flush previous gene */
			if (i > 0) {
				print_hdr (gname, gdir, mptr - mrna, mcds_begin, mcds_end);
				print_seq (gdir, mptr - mrna, mrna);
			}
			/* start new gene */
			gname = gene_name[i];
			gdir = dir[i];
			assert (++i < exlen);     /* cds line */
			gcds_begin = begin[i];
			gcds_end = end[i];
			mcds_begin = mcds_end = 0;
			mptr = mrna; *mptr = '\0';
		}
		else {     /* exon */
			xbeg = begin[i]; xend = end[i];

			/* check cds endpoints */
			if (xbeg <= gcds_begin && gcds_begin <= xend)
				mcds_begin = (mptr - mrna) + (gcds_begin - xbeg + 1);
			if (xbeg <= gcds_end && gcds_end <= xend)
				mcds_end = (mptr - mrna) + (gcds_end - xbeg + 1);

			/* append sequence fragment */
			if (xend > seqlen)
				fatalf ("Exon endpoint %d is beyond end of sequence.", xend);
			if ((mptr - mrna) + (xend - xbeg + 1) > BUFSIZE-1)
				fatalf ("mRNA for \"%s\" is too long for buffer.", gname);
			for (j = xbeg-1; j < xend; j++)   /* seq starts at 0 */
				*mptr++ = seq[j];
			*mptr = '\0';
		}
	}
	/* flush last gene */
	if (i > 0) {
		print_hdr (gname, gdir, mptr - mrna, mcds_begin, mcds_end);
		print_seq (gdir, mptr - mrna, mrna);
	}

	/* warn if additional (unprocessed) sequences */
	if (seq_read (sp) > 0) {
		fprintf (stderr, "Warning: %s does not support multiple contigs\n"
		         "in the sequence file -- only the first was used.\n", argv0);
	}

	/* clean up */
	free_exons (begin, end, dir, gene_name, exon_name, exlen);
	seq_close (sp);
	return 0;
}

static void get_args (int argc, char **argv, char **exonfile, char **seqfile) {
	int i;

	for (i = 1; i < argc; i++) {
		if (!*exonfile)
			*exonfile = argv[i];
		else if (!*seqfile)
			*seqfile = argv[i];
		else
			fatal (usage (SYNTAX));
	}
}

static void print_hdr (char *gname, int gdir, int length, int mcds_begin,
                       int mcds_end) {
	if (length > 0) {
		printf (">%s %d bp", gname, length);
		if (mcds_begin > 0 && mcds_end > 0) {
			if (gdir >= 0)
				printf (" CDS=%d..%d", mcds_begin, mcds_end);
			else
				printf (" CDS=%d..%d", length - mcds_end + 1,
				        length - mcds_begin + 1);
		}
		printf ("\n");
	}
	else
		fprintf (stderr, "Warning: skipping gene '%s' -- no exons.\n", gname);
}

static void print_seq (int gdir, int length, uchar *mrna) {
	uchar *old, *mptr;
	int cols;

	if (gdir < 0) {
		old = (uchar *) copy_string ((const char *) mrna);
		do_revcomp (mrna, length);
		if ((mptr = (uchar *) strrchr ((const char *) mrna, ' ')))
			fatalf ("Cannot compute complement of '%c'.",
			        old [length - ((mptr - mrna) + 1)]);
		free (old);
	}
	mptr = mrna;
	while (*mptr) {
		for (cols = 0; *mptr && cols < WIDTH; cols++)
			putchar (*mptr++);
		putchar ('\n');
	}
}
